#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include "sg_include.h"
#include "sg_err.h"

/* A utility program for the Linux OS SCSI generic ("sg") device driver.
*  Copyright (C) 2000-2003 D. Gilbert
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2, or (at your option)
*  any later version.

   This program outputs information provided by a SCSI MODE SENSE command.
   Does 10 byte MODE SENSE commands by default, Trent Piepho added a "-6"
   switch for force 6 byte mode sense commands.
   
*/

static char * version_str = "0.14 20030320";

#define ME "sg_modes: "

/* #define SG_DEBUG */

#define SENSE_BUFF_LEN 32       /* Arbitrary, could be larger */
#define DEF_TIMEOUT 60000       /* 60,000 millisecs == 60 seconds */

#define MODE_SENSE6_CMD      0x1a
#define MODE_SENSE6_CMDLEN   6
#define MODE_SENSE10_CMD     0x5a
#define MODE_SENSE10_CMDLEN  10
#define INQUIRY_CMD     0x12
#define INQUIRY_CMDLEN  6
#define MX_ALLOC_LEN (1024 * 4)

#define PG_CODE_ALL 0x3f

#define EBUFF_SZ 256


/* Returns 0 when successful, else -1 */
static int do_simple_inq(int sg_fd, void * resp, int mx_resp_len, int noisy)
{
    int res;
    unsigned char inqCmdBlk[INQUIRY_CMDLEN] = {INQUIRY_CMD, 0, 0, 0, 0, 0};
    unsigned char sense_b[SENSE_BUFF_LEN];
    sg_io_hdr_t io_hdr;

    inqCmdBlk[4] = (unsigned char)mx_resp_len;
    memset(&io_hdr, 0, sizeof(sg_io_hdr_t));
    io_hdr.interface_id = 'S';
    io_hdr.cmd_len = sizeof(inqCmdBlk);
    io_hdr.mx_sb_len = sizeof(sense_b);
    io_hdr.dxfer_direction = SG_DXFER_FROM_DEV;
    io_hdr.dxfer_len = mx_resp_len;
    io_hdr.dxferp = resp;
    io_hdr.cmdp = inqCmdBlk;
    io_hdr.sbp = sense_b;
    io_hdr.timeout = DEF_TIMEOUT;

    if (ioctl(sg_fd, SG_IO, &io_hdr) < 0) {
        perror("SG_IO (inquiry) error");
        return -1;
    }
    res = sg_err_category3(&io_hdr);
    switch (res) {
    case SG_ERR_CAT_CLEAN:
    case SG_ERR_CAT_RECOVERED:
	return 0;
    default:
	if (noisy) {
	    char ebuff[EBUFF_SZ];
	    snprintf(ebuff, EBUFF_SZ, "Inquiry error ");
            sg_chk_n_print3(ebuff, &io_hdr);
	}
	return -1;
    }
}

static int do_modes(int sg_fd, int dbd, int pc, int pg_code, 
		  void * resp, int mx_resp_len, int noisy, int mode6)
{
    int res;
    unsigned char modesCmdBlk[MODE_SENSE10_CMDLEN] = 
    	{MODE_SENSE10_CMD, 0, 0, 0, 0, 0, 0, 0, 0, 0};
    unsigned char sense_b[SENSE_BUFF_LEN];
    sg_io_hdr_t io_hdr;

    modesCmdBlk[1] = (unsigned char)(dbd ? 0x8 : 0);
    modesCmdBlk[2] = (unsigned char)(((pc << 6) & 0xc0) | (pg_code & 0x3f));
    if (mx_resp_len > (mode6?0xff:0xffff)) {
    	printf( ME "mx_resp_len too big\n");
	return -1;
    }
    if(mode6) {
	modesCmdBlk[0] = MODE_SENSE6_CMD;
	modesCmdBlk[4] = (unsigned char)(mx_resp_len & 0xff);
    } else {
	modesCmdBlk[7] = (unsigned char)((mx_resp_len >> 8) & 0xff);
	modesCmdBlk[8] = (unsigned char)(mx_resp_len & 0xff);
    }

    memset(&io_hdr, 0, sizeof(sg_io_hdr_t));
    io_hdr.interface_id = 'S';
    io_hdr.cmd_len = mode6 ? MODE_SENSE6_CMDLEN : MODE_SENSE10_CMDLEN;
    io_hdr.mx_sb_len = sizeof(sense_b);
    io_hdr.dxfer_direction = SG_DXFER_FROM_DEV;
    io_hdr.dxfer_len = mx_resp_len;
    io_hdr.dxferp = resp;
    io_hdr.cmdp = modesCmdBlk;
    io_hdr.sbp = sense_b;
    io_hdr.timeout = DEF_TIMEOUT;

    if (ioctl(sg_fd, SG_IO, &io_hdr) < 0) {
        perror("SG_IO (mode sense) error");
        return -1;
    }
    res = sg_err_category3(&io_hdr);
    switch (res) {
    case SG_ERR_CAT_CLEAN:
    case SG_ERR_CAT_RECOVERED:
	return 0;
    default:
	if (noisy) {
	    char ebuff[EBUFF_SZ];
	    snprintf(ebuff, EBUFF_SZ, "Mode sense error, dbd=%d, "
	    	     "pc=%d, page_code=%x ", dbd, pc, pg_code);
            sg_chk_n_print3(ebuff, &io_hdr);
	}
	return -1;
    }
}

struct page_code_desc {
	int page_code;
	const char * desc;
};
static struct page_code_desc pc_desc_arr[] = {
	{0x0, "vendor specific (does not require page format)"},
	{0x1, "Read-Write error recovery (direct access)"},
	{0x2, "Disconnect-Reconnect (all)"},
	{0x3, "Format (direct access))"},
	{0x4, "Rigid disk geometry (direct access)"},
	{0x5, "Flexible geometry (direct access)"},
	{0x7, "Verify error recovery (direct access)"},
	{0x8, "Caching (direct access)"},
	{0x9, "Peripheral device (direct access)"},
	{0xa, "Control (all)"},
	{0xb, "Medium types supported (direct access)"},
	{0xc, "Notch and partition (direct access)"},
	{0xf, "Compression (streaming: tape)"},
	{0x10, "XOR control (direct access)"},
	{0x11, "Partition (streaming: tape)"},
	{0x15, "Extended (all)"},
	{0x1a, "Power (all)"},
	{0x1c, "Informational exceptions (all)"},
	{0x3f, "[yields all supported pages]"},
};

const char * find_page_code_desc(int page_num)
{
    int k;
    int num = sizeof(pc_desc_arr) / sizeof(pc_desc_arr[0]);
    const struct page_code_desc * pcdp = &pc_desc_arr[0];

    for (k = 0; k < num; ++k, ++pcdp) {
	if (page_num == pcdp->page_code)
	    return pcdp->desc;
	else if (page_num < pcdp->page_code)
	    return NULL;
    }
    return NULL;
}

static void list_page_codes()
{
    int k;
    int num = sizeof(pc_desc_arr) / sizeof(pc_desc_arr[0]);
    const struct page_code_desc * pcdp = &pc_desc_arr[0];

    printf("Page_Code  Description\n");
    for (k = 0; k < num; ++k, ++pcdp)
	printf(" 0x%02x      %s\n", pcdp->page_code, pcdp->desc);   
}

static void usage()
{
    printf("Usage: 'sg_modes [-a] [-c=<page_control] [-d] [-h]"
	   " [-l] [-p=<page_number>]\n\t\t [-V] [-6] [<sg_device>]'\n"
	   " where -a   get all mode pages\n"
	   "       -c=<page_control> page control (def: 0 [current],"
	   " 1 [changeable],\n            2 [default], 3 [saved])\n"
	   "       -d   disable block descriptors\n"
	   "       -h   output in hex\n"
	   "       -l   list common page codes\n"
	   "       -p=<page_code> page code (in hex, def: 0)\n"
	   "       -V   output version string\n"
	   "       -6   Use MODE SENSE(6) instead of MODE SENSE(10)\n"
	   "       -?   output this usage message\n");
}


static void dStrHex(const char* str, int len, int no_ascii)
{
    const char* p = str;
    unsigned char c;
    char buff[82];
    int a = 0;
    const int bpstart = 5;
    const int cpstart = 60;
    int cpos = cpstart;
    int bpos = bpstart;
    int i, k;
    
    if (len <= 0) return;
    memset(buff,' ',80);
    buff[80]='\0';
    k = sprintf(buff + 1, "%.2x", a);
    buff[k + 1] = ' ';
    if (bpos >= ((bpstart + (9 * 3))))
    	bpos++;

    for(i = 0; i < len; i++)
    {
	c = *p++;
	bpos += 3;
	if (bpos == (bpstart + (9 * 3)))
	    bpos++;
	sprintf(&buff[bpos], "%.2x", (int)(unsigned char)c);
	buff[bpos + 2] = ' ';
	if (no_ascii)
	    buff[cpos++] = ' ';
	else {
	    if ((c < ' ') || (c >= 0x7f))
	        c='.';
	    buff[cpos++] = c;
	}
	if (cpos > (cpstart+15))
	{
	    printf("%s\n", buff);
	    bpos = bpstart;
	    cpos = cpstart;
	    a += 16;
	    memset(buff,' ',80);
	    k = sprintf(buff + 1, "%.2x", a);
	    buff[k + 1] = ' ';
	}
    }
    if (cpos > cpstart)
    {
	printf("%s\n", buff);
    }
}



int main(int argc, char * argv[])
{
    int sg_fd, k, num, len, md_len, bd_len, longlba, page_num;
    char * file_name = 0;
    char ebuff[EBUFF_SZ];
    const char * descp;
    unsigned char rsp_buff[MX_ALLOC_LEN];
    int rsp_buff_size = MX_ALLOC_LEN;
    unsigned int u;
    int pg_code = 0;
    int pc = 0;
    int do_all = 0;
    int do_dbd = 0;
    int do_hex = 0;
    int do_mode6 = 0;  /* Use MODE SENSE(6) instead of MODE SENSE(10) */
    int do_list = 0;
    int oflags = O_RDONLY;
    struct sg_scsi_id a_sid;
    int scsi_ptype, density_code_off;
    unsigned char * ucp;
    unsigned char uc;

    for (k = 1; k < argc; ++k) {
        if (0 == strncmp("-p=", argv[k], 3)) {
            num = sscanf(argv[k] + 3, "%x", &u);
            if ((1 != num) || (u > 63)) {
                printf("Bad page code after '-p' switch\n");
                file_name = 0;
                break;
            }
	    pg_code = u;
        }
        else if (0 == strncmp("-c=", argv[k], 3)) {
            num = sscanf(argv[k] + 3, "%x", &u);
            if ((1 != num) || (u > 3)) {
                printf("Bad page control after '-c' switch\n");
                file_name = 0;
                break;
            }
	    pc = u;
        }
        else if (0 == strcmp("-d", argv[k]))
	    do_dbd = 1;
        else if (0 == strcmp("-a", argv[k]))
	    do_all = 1;
        else if (0 == strcmp("-h", argv[k]))
	    do_hex = 1;
	else if (0 == strcmp("-6", argv[k]))
	    do_mode6 = 1;
	else if (0 == strcmp("-l", argv[k]))
	    do_list = 1;
        else if (0 == strcmp("-?", argv[k])) {
	    usage();
	    return 0;
	}
        else if (0 == strcmp("-V", argv[k])) {
	    printf("Version string: %s\n", version_str);
	    exit(0);
	}
        else if (*argv[k] == '-') {
            printf("Unrecognized switch: %s\n", argv[k]);
            file_name = 0;
            break;
        }
        else if (0 == file_name)
            file_name = argv[k];
        else {
            printf("too many arguments\n");
            file_name = 0;
            break;
        }
    }
    if (do_list) {
        list_page_codes();
	return 0;
    }
    else if (0 == file_name) {
        usage();
        return 1;
    }

    /* The 6 bytes command only allows up to 255 bytes of response data */
    if(do_mode6) rsp_buff_size = 255;

    if ((sg_fd = open(file_name, oflags)) < 0) {
        snprintf(ebuff, EBUFF_SZ, ME "error opening file: %s", file_name);
        perror(ebuff);
        return 1;
    }
    /* Just to be safe, check we have a new sg device by trying an ioctl */
    if ((ioctl(sg_fd, SG_GET_VERSION_NUM, &k) < 0) || (k < 30000)) {
        printf(ME "%s doesn't seem to be a version 3 sg device\n",
               file_name);
        close(sg_fd);
        return 1;
    }
    if (ioctl(sg_fd, SG_GET_SCSI_ID, &a_sid) < 0) {
	unsigned char inqBuff[36];

	if (do_simple_inq(sg_fd, inqBuff, sizeof(inqBuff), 1)) {
            printf(ME "%s doesn't respond to a SCSI INQUIRY\n", file_name);
            close(sg_fd);
            return 1;
	}
	scsi_ptype = inqBuff[0] & 0x1f; /* fetch peripheral device type */
    }
    else
	scsi_ptype = a_sid.scsi_type;
    printf("  SCSI peripheral type=0x%x (from INQUIRY)\n", scsi_ptype);
    if (do_all)
    	pg_code = PG_CODE_ALL;

    if (0 == do_modes(sg_fd, do_dbd, pc, pg_code, rsp_buff, rsp_buff_size, 1, 
                      do_mode6))
    {
	int medium_type, specific, headerlen;
    	printf("Mode parameter header:\n");
	if(do_mode6) {
	    headerlen = 4;
	    if (do_hex)
		dStrHex((const char *)rsp_buff, headerlen, 1);
	    md_len = rsp_buff[0]+1;
	    bd_len = rsp_buff[3];
	    medium_type = rsp_buff[1];
	    specific = rsp_buff[2];
	    longlba = 0; /* what is this field? */
	} else {
	    headerlen = 8;
	    md_len = (rsp_buff[0] << 8) + rsp_buff[1] + 2;
	    bd_len = (rsp_buff[6] << 8) + rsp_buff[7];
	    medium_type = rsp_buff[2];
	    specific = rsp_buff[3];
	    longlba = rsp_buff[4] & 1;
	}
	if (do_hex)
	    dStrHex((const char *)rsp_buff, headerlen, 1);
    	printf("  Mode data length=%d, medium type=0x%.2x, specific"
	       " param=0x%.2x, longlba=%d\n", md_len, medium_type, 
	       specific, longlba);
        if (md_len > rsp_buff_size) {
            printf("Only fetched %d bytes of response, truncate output\n",
                   rsp_buff_size);
            md_len = rsp_buff_size;
	    if (bd_len + headerlen > rsp_buff_size)
		bd_len = rsp_buff_size - headerlen;
        }
    	printf("  Block descriptor length=%d\n", bd_len);
	if (bd_len > 0) {
	    len = 8;
	    density_code_off = 0;
	    num = bd_len;
	    if (longlba) {
		printf("> longlba block descriptors:\n");
		len = 16;
		density_code_off = 8;
	    }
	    else if (0 == scsi_ptype) { 
		printf("> Direct access device block descriptors:\n");
		density_code_off = 4;
	    }
	    else
		printf("> General mode parameter block descriptors:\n");

	    ucp = rsp_buff + headerlen;
	    while (num > 0) {
		printf("   Density code=0x%x\n", *(ucp + density_code_off));
		dStrHex((const char *)ucp, len, 1);
		ucp += len;
		num -= len;
	    }
	    printf("\n");
	}
	ucp = rsp_buff + bd_len + headerlen;	/* start of mode page(s) */
	md_len -= bd_len + headerlen;		/* length of mode page(s) */
	while (md_len > 0) { /* got mode page(s) */
	    uc = *ucp;
	    page_num = ucp[0] & 0x3f;
	    descp = do_hex ? NULL : find_page_code_desc(page_num);
	    if (uc & 0x40) {
		len = (ucp[2] << 8) + ucp[3] + 4;
		if (do_hex)
		    printf(">> page_code=0x%x, subpage code=0x%x\n", 
		           page_num, ucp[1]);
		else
		    printf(">> page_code=0x%x [%s] , subpage code=0x%x\n", 
		           page_num, (descp ? descp: "vendor??"), ucp[1]);
	    }
	    else {
		len = ucp[1] + 2;
		if (do_hex)
		    printf(">> page_code=0x%x\n", page_num);
		else
		    printf(">> page_code=0x%x [%s]\n", 
		           page_num, (descp ? descp: "vendor??"));
	    }
	    dStrHex((const char *)ucp, len, 1);
	    ucp += len;
	    md_len -= len;
	}
    }

    close(sg_fd);
    return 0;
}
